from typing import Dict, Union, Any, Optional, Type

from extra_qt.dom.qt_dom import create_element
from extra_qt.virtual_dom import ChildrenT, VirtualNode


class Component:
    props: Dict[Union[int, str], Any]
    children: ChildrenT

    wrapper: Type['ComponentWrapper'] = None

    initial_state: Optional[Any] = None
    state: Optional[Any] = None

    def __init__(self, props, children):
        self.props = props
        self.children = children

        self.state = self.get_initial_state()

    @classmethod
    def c(cls, *args, **kwargs):
        return create_element(cls, *args, **kwargs)

    @classmethod
    def get_initial_state(cls):
        return cls.initial_state

    @staticmethod
    def updates_state(state_method):
        def perform_set_state(self):
            def call_state_method(s):
                bound_method = getattr(s, state_method.__name__)
                bound_method()
                return s

            self.set_state(call_state_method)

        return perform_set_state

    def set_state(self, latest_state):
        from extra_qt.reconciler import reconciler
        self.wrapper.pending_state.append(latest_state)
        reconciler.update_if_necessary(self.wrapper)

    def render(self) -> VirtualNode:
        raise NotImplementedError()

    def before_receive_props(self, next_props, next_children):
        pass

    def should_update(self, next_props, next_children, next_state):
        return True

    def before_mount(self):
        pass

    def after_mount(self):
        pass

    def before_unmount(self):
        pass




from typing import Type, Union

from extra_qt.virtual_dom import TagType, VirtualNode

__all__ = (
    'create_element',
    'Label', 'label',
    'Button', 'button',
    'Group', 'group',
    'Tabs', 'tabs',
    'CheckBox', 'check_box',
    'ComboBox', 'combo_box',
    'SpinBox', 'spin_box',
    'LineEdit', 'line_edit',
    'Slider', 'slider',
    'TextEdit', 'text_edit',
    'Dial', 'dial',
)


def create_element(tag_type: Union[TagType, Type['Component']], props=None, children=None) -> VirtualNode:
    if isinstance(props, (list, str)):
        children = props
        props = {}

    if props is None:
        props = dict()

    return VirtualNode(tag_type, props=props, children=children or [])


def _bind_create(tag):
    def _create_element(*args, **kwargs):
        return create_element(tag, *args, **kwargs)

    return _create_element


def _bind_create_input(tag):
    def _create_element(**props):
        return create_element(tag, props=props)

    return _create_element


Label = TagType.LABEL
Button = TagType.BUTTON
Group = TagType.GROUP
Tabs = TagType.TABS
CheckBox = TagType.CHECK_BOX
ComboBox = TagType.COMBO_BOX
SpinBox = TagType.SPIN_BOX
LineEdit = TagType.LINE_EDIT
Dial = TagType.DIAL
Slider = TagType.SLIDER
TextEdit = TagType.TEXT_EDIT

button = _bind_create_input(Button)
group = _bind_create(Group)
tabs = _bind_create(Tabs)
check_box = _bind_create_input(CheckBox)
combo_box = _bind_create_input(ComboBox)
spin_box = _bind_create_input(SpinBox)
line_edit = _bind_create_input(LineEdit)
dial = _bind_create_input(Dial)
slider = _bind_create_input(Slider)
text_edit = _bind_create_input(TextEdit)


def label(text, props=None, children=None):
    props = props or {}
    props['text'] = text
    return create_element(Label, props, [])


"""
Essentially a tiny experimental port of React to Python to use Qt rendering.
"""

__version__ = '0.1.0'


from PyQt5.QtWidgets import QWidget

from extra_qt.renderers.renderer import ComponentWrapper, WrapperT
from extra_qt.virtual_dom import VirtualNode, TagType
from extra_qt.component import Component
from extra_qt.reconciler import reconciler


def initial_render(element: VirtualNode, container: QWidget) -> QWidget:
    instance = reconciler.wrap(element)
    widget = reconciler.mount(instance, container)
    container.rendered = instance

    return widget


def render(element: VirtualNode, container: QWidget):
    try:
        prev_component: WrapperT = container.rendered
    except AttributeError:
        return initial_render(element, container)

    return update_from_root(prev_component, element)


def update_from_root(previous: WrapperT, element: VirtualNode):
    return reconciler.receive(previous, element)





from typing import Type, Any

from extra_qt.virtual_dom import VirtualNode, TagType

__all__ = ('reconciler',)

WrapperT = Type['WrapperT']
HostNode = Any


class Reconciler:
    host_wrapper_cls: Type['HostWrapper'] = None
    host_node_cls: Any = None

    def configure(self, host_wrapper_cls=None, host_node_cls=None):
        from .renderers.qt_renderer import QWidget, QtDOMWrapper
        self.host_wrapper_cls = host_wrapper_cls or QtDOMWrapper
        self.host_node_cls = host_node_cls or QWidget

    def wrap(self, element) -> WrapperT:
        from extra_qt.renderers.renderer import ComponentWrapper
        if isinstance(element.tag_type, TagType):
            return self.host_wrapper_cls(element)

        return ComponentWrapper(element)

    @staticmethod
    def mount(instance: WrapperT, container: HostNode) -> HostNode:
        return instance.mount(container)

    @staticmethod
    def unmount(instance: WrapperT, container: HostNode = None):
        return instance.unmount(container, container)

    @staticmethod
    def receive(instance: WrapperT, latest: VirtualNode) -> HostNode:
        instance.receive(latest)

    @staticmethod
    def update_if_necessary(instance: WrapperT):
        instance.update_if_necessary()


reconciler = Reconciler()


from dataclasses import dataclass, field
from enum import Enum
from typing import List, Any, Dict, Type
import itertools

import typing


@dataclass
class VirtualStyleSettings:
    pass


class TagType(Enum):
    LABEL = 1
    GROUP = 2

    BUTTON = 4

    # partially supported
    TABS = 3

    CHECK_BOX = 5
    COMBO_BOX = 6
    SPIN_BOX = 7

    LINE_EDIT = 8

    DIAL = 9
    SLIDER = 10
    TEXT_EDIT = 11

    # FUTURE, unsupported
    # ========================
    # structural

    # input
    TUMBLER = 12
    FILE_DIALOG = 13


class LayoutType(Enum):
    VERTICAL = 0
    HORIZONTAL = 1
    GRID = 2


ChildrenT = typing.Union['VirtualNode', List['VirtualNode']]


@dataclass
class VirtualNode:
    tag_type: typing.Union[TagType, Type['Component']] = TagType.LABEL
    children: ChildrenT = field(default_factory=list)
    props: Dict[str, Any] = field(default_factory=dict)

    def repr_tree(self):
        return '\n'.join(self.repr_tree_node())

    def repr_tree_node(self) -> List[str]:
        lines = [
            f'{self.tag_type.name if isinstance(self.tag_type, TagType) else self.tag_type.__name__}',
        ]

        def safe_repr_node(node_or_str):
            return node_or_str if isinstance(node_or_str, str) else node_or_str.repr_tree_node()

        if isinstance(self.children, str):
            child_lines = [[self.children]]
        elif isinstance(self.children, list):
            child_lines = [safe_repr_node(child) for child in self.children]
        else:
            child_lines = [[safe_repr_node(self.children)]]

        return lines + ['  ' + l for l in itertools.chain(*child_lines)]



from dataclasses import dataclass, field
from typing import Union, List, Any, Type

from PyQt5.QtWidgets import QWidget

from extra_qt.component import Component
from extra_qt.virtual_dom import VirtualNode
from extra_qt.reconciler import reconciler

__all__ = ('ComponentWrapper', 'HostWrapper', 'WrapperT')

WrapperT = Union['ComponentWrapper', 'HostWrapper']
HostNodeT = Any

Updateable = (dict,)


@dataclass
class ComponentWrapper:
    element: VirtualNode  # virtual markup for this component
    host_container: Any = None # Where the underlying node is mounted
    component: Component = None  # Client Component with lifecycle

    # We hold exactly one child, a wrapper for the contents of ``self.component.render()``
    # this may be a further component, or a host DOM tree.
    wrapped_child: WrapperT = None

    pending_state: List[Any] = field(default_factory=list)

    is_rendering: bool = False

    def receive(self, element: VirtualNode):
        return self.update(self.element, element)

    @property
    def next_state(self):
        latest_state = self.component.state

        for state_update in self.pending_state:
            if isinstance(latest_state, Updateable) and isinstance(state_update, Updateable):
                latest_state.update(state_update)
            elif callable(state_update):
                latest_state = state_update(latest_state)
            else:
                latest_state = state_update

        self.pending_state = []
        return latest_state

    def update(self, previous: VirtualNode, latest: VirtualNode):
        self.is_rendering = True
        if previous != latest:
            self.component.before_receive_props(latest.props, latest.children)

        latest_state = self.next_state
        self.component.props = latest.props
        self.component.children = latest.children
        self.component.state = latest_state

        if self.component.should_update(latest.props, latest.children, latest_state):
            self.element = latest
            self.update_child()

        self.is_rendering = False

    def update_child(self):
        new_element = self.component.render()

        if new_element.tag_type == self.wrapped_child.element.tag_type:
            reconciler.receive(self.wrapped_child, new_element)
        else:
            # uproot our contents and remount onto the parent container
            reconciler.unmount(self.wrapped_child)
            self.wrapped_child = reconciler.wrap(new_element)
            reconciler.mount(self.wrapped_child, self.host_container)

    def mount(self, container: QWidget) -> QWidget:
        component_cls: Type[Component] = self.element.tag_type
        self.component = component_cls(self.element.props, self.element.children)
        self.component.wrapper = self

        self.component.before_mount()
        widget = self.initial_mount(container)
        self.component.after_mount()

        return widget

    def unmount(self, container: QWidget = None):
        self.component.before_unmount()
        reconciler.unmount(self.wrapped_child, container)

    def initial_mount(self, container: reconciler.host_wrapper_cls):
        self.host_container = container
        self.wrapped_child = reconciler.wrap(self.component.render())
        return reconciler.mount(self.wrapped_child, container)

    def update_if_necessary(self):
        # don't rerender if we are already rendering.
        if not self.is_rendering:
            self.update(self.element, self.element)


class HostWrapper:
    """
    Essentially defines the same interface as ReactDOMComponent (not React.Component!)
    Because of this frankly, confusing, terminology in React I call these wrappers here.

    You can subclass this to provide custom renderers. Currently we are working towards
    supporting:

      1. Qt5
      2. Dict-literal (for testing)

    There are a few other pieces of code that may need adjusting depending on the renderer.
    In particular, the virtual DOM we use has typed tags (Enum<int>). This is a very simple fix
    if necessary, and for now I like the control that the Enum provides as opposed to bare
    strings.
    """

    @classmethod
    def use_as_renderer(cls):
        """
        Configure the reconciler to use Qt as the render target. This amounts currently
        just to telling the reconciler what it is supposed to wrap host elements
        (i.e. with ``type(el.tag_type) =inst= TagType``) inside.
        """
        from extra_qt import reconciler
        reconciler.host_wrapper_cls = cls
        reconciler.host_node_cls = QWidget

    element: VirtualNode  # virtual markup for this component
    host_node: HostNodeT = None  # what we rendered to
    host_container: HostNodeT = None  # where the node we rendered to is attached
    wrapped_children: List[WrapperT] = field(default_factory=list)

    def receive(self, element: VirtualNode):
        """
        Called when the wrapper receives new markup in order to trigger an update.
        What happens here depends a lot currently on the render target, i.e. that we are
        targeting Qt, as a result, this is implemented (for instance) in ``.qt_renderer``
        """
        raise NotImplementedError()

    def unmount(self, container: HostNodeT = None):
        """
        Uproot me from the rendered tree. This is highly dependent upon the render target.
        For instance, for the Qt renderer, this amounts to disposing of all downstream
        (children, their children, ...) widgets, removing them, and removing
        ``self.hode_node`` from ``self.host_container`` with

        ``self.host_container.layout().remove_widget(self.host_node)``
        """
        raise NotImplementedError()

    def mount(self, container: HostNodeT):
        """
        Attach me to the render tree. This consists of generating (rendering) the output,
        be it a QWidget, a dictionary, or something else, before ultimately attaching it onto
        the target container.
        """
        raise NotImplementedError()



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Dict, Callable, Optional, Union, Tuple, Any

from extra_qt import render
from extra_qt.reconciler import reconciler
from extra_qt.virtual_dom import VirtualNode, TagType
from .renderer import HostWrapper, WrapperT

from PyQt5.QtWidgets import QWidget, QLabel, QVBoxLayout, QGroupBox, QPushButton, QMainWindow, QApplication, QTabWidget, \
    QCheckBox, QSpinBox, QLineEdit, QDial, QSlider, QTextEdit, QComboBox

__all__ = ('QtDOMWrapper', 'render_window', 'MarkupFormat',)


class MarkupFormat(Enum):
    TEXT = 0
    HTML = 1
    MARKDOWN = 2


def set_style(w: QWidget, style: any):
    if style is not None:
        if not isinstance(style, str):
            w.setStyleSheet(style.to_stylesheet())
        else:
            w.setStyleSheet(style)


def update_style(w: QWidget, previous_style: any, next_style: any):
    if previous_style == next_style:
        return

    set_style(w, next_style or '{}')


def noop(*_, **__):
    pass


def noop_build(cls):
    def builder(_props, _children):
        return cls()

    return builder


def set_widget_style_and_signals(**signal_map: Dict[str, Union[Tuple[Any, str], str]]):
    """
    Example of how signals and slots are treated:

    @set_..._signals(activated=(str, 'on_change')) <->
         widget.activated[str].connect(props['on_change'])

    @set_..._signals(pressed='on_click')) <->
         widget.pressed.connect(props['on_click'])

    with some extra safety. This is also the behavior on the update decorator.
    """
    signal_map: Dict[str, Tuple[Any, str]] = {
        signal_name: (None, slot_name) if isinstance(slot_name, str) else slot_name
        for signal_name, slot_name in signal_map.items()
    }

    def decorate(build):
        if isinstance(build, type):
            build = noop_build(build)

        def wrapped_build(props, children):
            w = build(props, children)
            set_style(w, props.get('style'))

            for signal_name, (arg, handler_name) in signal_map.items():
                if handler_name in props:
                    signal = getattr(w, signal_name)

                    if arg is not None:
                        signal = signal[arg]

                    signal.connect(props[handler_name])

            return w

        return wrapped_build
    return decorate


def update_widget_style_and_signals(**signal_map):
    signal_map = {
        signal_name: (None, slot_name) if isinstance(slot_name, str) else slot_name
        for signal_name, slot_name in signal_map.items()
    }

    def decorate(updater=None):
        if updater is None:
            updater = noop

        def wrapped_update(w: QWidget, previous: VirtualNode, latest: VirtualNode):
            updater(w, previous, latest)
            update_style(w, previous.props.get('style'), latest.props.get('style'))

            for signal_name, (arg, handler_name) in signal_map.items():
                p_handler, l_handler = previous.props.get(handler_name), latest.props.get(handler_name)
                if p_handler != l_handler:
                    signal = getattr(w, signal_name)
                    if arg is not None:
                        signal = signal[arg]

                    if p_handler:
                        signal.disconnect(p_handler)

                    signal.connect(l_handler)

        return wrapped_update

    return decorate


@set_widget_style_and_signals()
def build_label(props, children):
    return QLabel(props.get('text', ''))


@set_widget_style_and_signals()
def build_group(props, children):
    w = QGroupBox(props.get('title', ''))
    w.setLayout(QVBoxLayout())
    return w


@set_widget_style_and_signals(pressed='on_click')
def build_button(props, children):
    return QPushButton(props.get('text', ''))


def build_tabs(props, children):
    w = QTabWidget()

    labels = props.get('labels', [])
    for p in labels:
        tab = QWidget()
        tab.setLayout(QVBoxLayout())
        w.addTab(tab, p)

    selected = props.get('selected')
    if isinstance(selected, str):
        selected = labels.index(selected)

    if selected is not None:
        w.setCurrentIndex(selected)

    return w


@set_widget_style_and_signals(stateChanged='on_change')
def build_check_box(props, children):
    return QCheckBox(props.get('title', ''))


@set_widget_style_and_signals(activated=(str, 'on_change'))
def build_combo_box(props, children):
    w = QComboBox()
    keys = props.get('keys')
    values = props.get('values', [None] * len(keys))

    for k, v in zip(keys, values):
        w.addItem(k, v)

    return w


@set_widget_style_and_signals()
def build_text_edit(props, children):
    w = QTextEdit()
    handler = props.get('on_change')
    format = props.get('format', MarkupFormat.TEXT)

    if handler:
        def wrapped_handler():
            if format == MarkupFormat.TEXT:
                handler(w.toPlainText())
            elif format == MarkupFormat.HTML:
                handler(w.toHtml())
            elif format == MarkupFormat.MARKDOWN:
                handler(w.toMarkdown())

        w.textChanged.connect(wrapped_handler)

    return w


component_tag_map: Dict[TagType, Callable[[], QWidget]] = {
    TagType.LABEL: build_label,
    TagType.GROUP: build_group,
    TagType.BUTTON: build_button,
    TagType.TABS: build_tabs,
    TagType.CHECK_BOX: build_check_box,
    TagType.COMBO_BOX: build_combo_box,
    TagType.SPIN_BOX: set_widget_style_and_signals(valueChanged='on_change')(QSpinBox),
    TagType.LINE_EDIT: set_widget_style_and_signals(textChanged='on_change')(QLineEdit),
    TagType.DIAL: set_widget_style_and_signals(valueChanged='on_change')(QDial),
    TagType.SLIDER: set_widget_style_and_signals(valueChanged='on_change')(QSlider),
    TagType.TEXT_EDIT: build_text_edit,
}


@update_widget_style_and_signals()
def update_label(w: QLabel, previous: VirtualNode, latest: VirtualNode):
    w.setText(latest.props.get('text', ''))


@update_widget_style_and_signals()
def update_group(w: QGroupBox, previous: VirtualNode, latest: VirtualNode):
    w.setTitle(latest.props.get('title', ''))


@update_widget_style_and_signals(pressed='on_click')
def update_button(w: QPushButton, previous: VirtualNode, latest: VirtualNode):
    w.setText(latest.props.get('text', ''))


def update_tabs(w: QTabWidget, previous: VirtualNode, latest: VirtualNode):
    p_selected, l_selected = previous.props.get('selected'), latest.props.get('selected')

    if isinstance(p_selected, str):
        p_selected = previous.props.get('labels', []).index(p_selected)

    if isinstance(l_selected, str):
        l_selected = latest.props.get('labels', []).index(l_selected)

    if p_selected != l_selected:
        if l_selected is not None:
            w.setCurrentIndex(l_selected)


@update_widget_style_and_signals()
def update_text_edit(w: QTextEdit, previous: VirtualNode, latest: VirtualNode):
    """
    We don't use textChanged='on_change' in the signal updater here because in Qt
    QTextEdit.textChanged is just a notifier, it does not get the text.

    If this is not performant enough we will have to deal with implementing refs.
    """
    p_handler, l_handler = previous.props.get('on_change'), latest.props.get('on_change')
    p_format, l_format = previous.props.get('format', MarkupFormat.TEXT), \
                         latest.props.get('format', MarkupFormat.TEXT)
    if p_handler != l_handler or p_format != l_format:
        w.textChanged.disconnect()  # disconnect all because we manage the handler here

        if l_handler is not None:
            def wrapped_handler():
                if l_format == MarkupFormat.TEXT:
                    l_handler(w.plainText)
                elif l_format == MarkupFormat.HTML:
                    l_handler(w.html)
                elif l_format == MarkupFormat.MARKDOWN:
                    l_handler(w.markdown)

            w.textChanged.connect(wrapped_handler)


tag_update_map: Dict[TagType, Callable[[QWidget, VirtualNode], None]] = {
    TagType.LABEL: update_label,
    TagType.GROUP: update_group,
    TagType.BUTTON: update_button,
    TagType.TABS: update_tabs,  # TODO, handle changes in number and order of tabs
    TagType.CHECK_BOX: update_widget_style_and_signals(stateChanged='on_change')(),
    TagType.COMBO_BOX: update_widget_style_and_signals(activated=(str, 'on_change'))(),
    TagType.SPIN_BOX: update_widget_style_and_signals(valueChanged='on_change')(),
    TagType.LINE_EDIT: update_widget_style_and_signals(textChanged='on_change')(),
    TagType.SLIDER: update_widget_style_and_signals(valueChanged='on_change')(),
    TagType.DIAL: update_widget_style_and_signals(valueChanged='on_change')(),
    TagType.TEXT_EDIT: update_text_edit,
}


@dataclass
class QtDOMWrapper(HostWrapper):
    element: VirtualNode

    host_node: QWidget = None
    host_container: QWidget = None
    wrapped_children: List[WrapperT] = field(default_factory=list)

    def receive(self, element: VirtualNode) -> QWidget:
        return self.update(self.element, element)

    def update(self, previous: VirtualNode, latest: VirtualNode) -> QWidget:
        self.update_properties(previous, latest)

        if self.element.tag_type not in {TagType.BUTTON, TagType.LABEL, }:
            self.update_children(previous, latest)

        self.element = latest

        return self.host_node

    def update_properties(self, previous: VirtualNode, latest: VirtualNode):
        tag_update_map[self.element.tag_type](self.host_node, previous, latest)

    def update_children(self, previous: VirtualNode, latest: VirtualNode):
        """
        Update the children for an actual node. This is naive because it does not
        handle moving children that do not need to be unmounted. This will need
        to be revisited once it becomes a problem.

        This also doesn't handle Tabs appropriately at the moment.
        """
        p_children = previous.children
        p_children: List[VirtualNode] = [p_children] if isinstance(p_children, VirtualNode) \
            else p_children
        l_children = latest.children
        l_children: List[VirtualNode] = [l_children] if isinstance(l_children, VirtualNode) \
            else l_children

        new_wrapped_children = []

        for i, child_elem in enumerate(l_children):
            try:
                previous_wrapped_child = self.wrapped_children[i]
            except IndexError:
                new_wrapped_child = reconciler.wrap(child_elem)
                reconciler.mount(new_wrapped_child, self.host_node)
                new_wrapped_children.append(new_wrapped_child)
                continue

            p_child, l_child = p_children[i], l_children[i]
            if p_child.tag_type != l_child.tag_type:
                # completely unmount the tree
                previous_wrapped_child.unmount()
                next_child = reconciler.wrap(l_child)
                reconciler.mount(next_child, self.host_node)
            else:
                reconciler.receive(previous_wrapped_child, l_child)
                new_wrapped_children.append(previous_wrapped_child)

        for stale_child in self.wrapped_children[len(l_children):]:
            reconciler.unmount(stale_child)

        self.wrapped_children = new_wrapped_children

    def set_text_content(self, text: str):
        label: QLabel = self.host_node
        label.setText(text)

    def inflate(self) -> QWidget:
        dom_element = component_tag_map[self.element.tag_type](self.element.props, self.element.children)
        return dom_element

    def unmount(self, container: QWidget = None):
        if container is None:
            container = self.host_container

        for i, wrapper in enumerate(self.wrapped_children):
            if self.element.tag_type == TagType.TABS:
                reconciler.unmount(wrapper, self.host_node.widget(i))
            else:
                reconciler.unmount(wrapper, self.host_node)

        container.layout().removeWidget(self.host_node)

    def mount(self, container: QWidget):
        self.host_container = container

        dom_element = self.inflate()
        container.layout().addWidget(dom_element)
        self.host_node = dom_element

        children = self.element.children
        if not isinstance(children, str):
            children = (children,) if isinstance(children, VirtualNode) else children
            for i, child in enumerate(children):
                wrapper = reconciler.wrap(child)
                self.wrapped_children.append(wrapper)

                if self.element.tag_type == TagType.TABS:
                    reconciler.mount(wrapper, dom_element.widget(i))
                else:
                    reconciler.mount(wrapper, dom_element)

        return dom_element

    @classmethod
    def use_as_renderer(cls):
        from extra_qt import reconciler
        reconciler.host_wrapper_cls = cls
        reconciler.host_node_cls = QWidget


def render_window(element: VirtualNode, window=None, after_show: Optional[Callable[[QMainWindow], None]] = None):
    old_window = window
    window = build_app_window() if window is None else window
    container = window.centralWidget()

    render(element, container)

    if not old_window:
        window.show()

    try:
        after_show(window)
    except TypeError:
        pass

    if not old_window:
        window.app.exec_()


def build_app_window():
    app = QApplication([])
    window = QMainWindow()
    screen_rect = app.primaryScreen().geometry()
    window.move(screen_rect.left(), screen_rect.top())
    window.resize(600, 600)
    w = QWidget()
    w.setLayout(QVBoxLayout())
    window.setCentralWidget(w)
    window.app = app
    return window